/*
-----------------------------------------------------
   Application Name: "DIMG"
Application Version: 1.0.0
          File Name: Dimg.cpp
         Start Date: 27/06/2008
       Last Changed: $LastChangedDate: 2011-06-12 16:41:13 +1200 (Sun, 12 Jun 2011) $
           Revision: $Revision: 310 $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2008, 2009
            License: GNU General Public License
   File Description: DIMG Class Implementation
-----------------------------------------------------
	This file is part of "DIMG".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

	File Notes:
	------------------
*/



// ------------------
// Header files:
// ------------------
#include "Dimg.h"
#include <QScopedPointer>

// ------------------
// Constructor for download_manager class
Dimg::Dimg(QApplication *app) : QMainWindow(0)
{
        qDebug() << "Dimg Constructor...";

        // Remember the pointer to the application
	pApp = app;

        currentImage = 0;
        imageDevice = 0;
        fs = 0;
        imgr = 0;

        // Image doesn't need saving
        markImageSaved();

	// Then setup the Ui
	ui.setupUi(this);

        volumeLabel = new QLabel(0);
        statusBar()->addWidget(volumeLabel);


        // Load the settings
        QString settingsFile;
        settingsFile = QCoreApplication::applicationName() + ".ini";
        qDebug() << "Checking for configuration file " << settingsFile;

        if (QFile::exists(settingsFile)) {
            // Load settings from there
            settings.reset(new QSettings(settingsFile, QSettings::IniFormat, 0));
        } else {
            settings.reset(new QSettings());
        }
        qDebug() << "Settings Loaded.";


        // Connect up a bunch of actions
        connect(ui.action_Open,SIGNAL(triggered()),this,SLOT(loadFile()));
        connect(ui.actionExtract_File,SIGNAL(triggered()),this,SLOT(extractFile()));
        connect(ui.actionFile_Properties,SIGNAL(triggered()),this,SLOT(showFileProperties()));
        connect(ui.actionE_xit,SIGNAL(triggered()),this,SLOT(close()));
        connect(ui.action_Close_Image,SIGNAL(triggered()),this,SLOT(closeImage()));
        connect(ui.action_About,SIGNAL(triggered()),this,SLOT(showAbout()));
        connect(ui.action_Read_Disk,SIGNAL(triggered()),this,SLOT(imageCurrentDrive()));
        connect(ui.actionWrite_Disk,SIGNAL(triggered()),this,SLOT(writeCurrentDrive()));
        connect(ui.action_Save,SIGNAL(triggered()),this,SLOT(saveCurrentImage()));
        connect(ui.actionSave_As,SIGNAL(triggered()),this,SLOT(saveCurrentImageAs()));
        connect(ui.action_ImageInformation,SIGNAL(triggered()),this,SLOT(showImageInformation()));
        connect(ui.actionS_ettings,SIGNAL(triggered()),this,SLOT(showSettings()));

        // tree widget
        connect(ui.treeWidget,SIGNAL(collapsed(QModelIndex)),this,SLOT(resize_columns(QModelIndex)));
        connect(ui.treeWidget,SIGNAL(expanded(QModelIndex)),this,SLOT(resize_columns(QModelIndex)));

        // ----------------
        // Load the Filesystem Plugin
        loadFSPlugin();

        // ----------------
        // Prepare the Imager
        if (!imgr) {
            imgr = new Imager(this);
            imgr->set_parent_widget(this);
        }

        // Load the disk drivers
        if (imgr->load_drivers() == errors::E_SUCCESS) {
            qDebug() << "Preparing drive selector...";
            // ----------------
            // Add the floppy drives to the Disk menu
            QStringList fdns = imgr->getDescriptiveDriveNames();
            QStringList fds = imgr->getDriveNames();

            if (!fds.isEmpty()) {
                QPointer<QMenu> diskMenu = ui.menu_Select_Disk;
                QMenu targetMenu;
                targetMenu.setTitle("Select Drive");
                QPointer<QToolBar> disksel = ui.tb_DiskSelect;

                for (int i = 0; i < fds.count(); i++) {
                        QString title = "Use drive " + fdns.at(i);
                        QVariant drive = fds.at(i);
                        QAction *tempAction = new QAction(title, this);
                        tempAction->setData(drive);
                        tempAction->setCheckable(true);
                        disksel->addAction(tempAction);
                        diskMenu->addAction(tempAction);
                        diskSelActions << tempAction;
                        qDebug() << "Added Drive Action: " << title;
                }
                connect(diskMenu,SIGNAL(triggered(QAction *)),this,SLOT(diskSelect(QAction *)));
                connect(disksel,SIGNAL(actionTriggered(QAction *)),this,SLOT(diskSelect(QAction *)));
            } else {
                ui.menu_Select_Disk->setEnabled(false);
                ui.action_Read_Disk->setEnabled(false);
                ui.actionWrite_Disk->setEnabled(false);
                ui.action_Format_Disk->setEnabled(false);
            }

            if (imgr->configurable()) {
                qDebug() << "Driver is configurable. Adding configure option";
                QAction *dconf = new QAction("Configure Driver",this);

                ui.menu_Disk->addSeparator();
                ui.menu_Disk->addAction(dconf);
                connect(dconf,SIGNAL(triggered()),imgr,SLOT(configure()));
            }
        }
        else {
            // Disable features that rely on a DQPlugin if one couldnt be loaded.
            ui.action_Read_Disk->setEnabled(false);
            ui.actionWrite_Disk->setEnabled(false);
            ui.action_Format_Disk->setEnabled(false);
        }

        // Disable the write disk button as we have no image loaded.
        ui.actionWrite_Disk->setEnabled(false);
        ui.action_ImageInformation->setEnabled(false);

        // Set the window title
        setProgramWindowTitle("");

        // Prepare the Filesystem Treeview widget
        prepareFSTreeView();

        // Prepare the image file handling stuff
        ifu.setSettings(settings.data());
        ifu.initialise();

        // If nothing is open the filesystem stuff should be disabled.
        setFSEnabled(false);

	// Show the user interface.
	show();
}

Dimg::~Dimg() {
    qDebug() << "Dimg Destructor...";
    if (fs) {
        fs->close();
        fs->exiting();
        delete fs;
        fs = 0;
    }
    if (currentImage) {
        currentImage->close();
        delete currentImage;
        currentImage = 0;
    }
}

bool Dimg::loadFSPlugin() {
    qDebug() << "Loading FS Plugin...";
    QString plugin_file, alt_file;

    // Set the directory for the plugin to be the applications directory
    plugin_file = alt_file = QApplication::applicationDirPath();

    // Build the base filename
#ifdef Q_OS_WIN32
    plugin_file += "/fat12fs";
    alt_file += "/fsplugin";
#else
    plugin_file += "/libfat12fs";
    alt_file += "/libfsplugin";
#endif

    // If it is a debug build, use the debug build of the plugin
#ifdef _DIMG_DEBUG
    plugin_file += "d";
    alt_file += "d";
#endif

    // Add on the extension
#ifdef Q_OS_WIN32
    plugin_file += ".dll";
    alt_file += ".dll";
#else
    plugin_file += ".so";
    alt_file += ".so";
#endif

    QFileInfo info(plugin_file);

    if (!info.exists()) {
        plugin_file = alt_file;
        info.setFile(plugin_file);
        if (!info.exists()) {
            plugin_file = QFileDialog::getOpenFileName(this,"Locate Filesystem Plugin");
            if (plugin_file.isEmpty()) {
                QMessageBox::critical(0,"Plugin Load Error","No filesystem plugin specified. Exiting.");
            }
        }
    }

    qDebug() << "Creating plugin loader for file " << plugin_file;
    QPluginLoader *loader = new QPluginLoader(plugin_file);
    qDebug() << "Loading plugin...";
    if (loader->load()) {
        fs = qobject_cast<FSPluginv2 *>(loader->instance());
        if (!fs) {
            QMessageBox::critical(0,"Plugin Load Error","Loaded plugin was not of type FSPlugin.");
        }
    }
    else {
        QMessageBox::critical(0,"Plugin Load Error","A filesystem plugin could be found but not loaded.\nThis is likely a problem with the plugin.\nThe Plugin filename is:\n" + plugin_file);
    }
    delete loader;
    loader = 0;
    return (fs);
}

void Dimg::prepareFSTreeView() {
    qDebug() << "Preparing Filesystem Tree View...";
    // If there is no filesystem plugin loaded, disable the tree widget and
    // return.
    if (!fs) {
        ui.treeWidget->setDisabled(true);
        return;
    }

    // Otherwise, setup the column labels. This is where we would add any
    // filesystem-specific column headers.

    // Build the column labels
    QStringList cl;
    cl << "Filename";
    cl << "Attributes";
    cl << "Size";
    cl << "Date Modified";

    ui.treeWidget->setColumnCount(cl.size());
    ui.treeWidget->setHeaderLabels(cl);

    // Setup context menu
    ui.treeWidget->addAction(ui.actionExtract_File);
    ui.treeWidget->addAction(ui.actionFile_Properties);
    ui.treeWidget->setContextMenuPolicy(Qt::ActionsContextMenu);
}

void Dimg::diskSelect(QAction * dskAction) {
        qDebug() << "Disk Selected: " << dskAction->data();
	for (int i = 0; i < diskSelActions.count(); i++)
		diskSelActions.at(i)->setChecked(false);
	dskAction->setChecked(true);

        if (imgr)
            imgr->set_current_drive(dskAction->data().toString());
}

void Dimg::loadFile() {
    qDebug() << "Load file requested";
    loadFile("");
}

void Dimg::loadFile(QString filename) {
    qDebug() << "Load file requested";

    ImageFile *imgfile = 0;
    int retval;
    if (filename.isEmpty()) {
        qDebug() << "No filename supplied. Prompting user.";
        retval = ifu.openImage(&imgfile);
    }
    else {
        qDebug() << "Attempting to load file " << filename;
        retval = ifu.openImage(&imgfile, filename);
    }

    /*
     * Was there success? was an ImageFile created? was it opened?
     */
    if (retval == errors::E_SUCCESS) {

        // Close any currently open image.
        closeImage();

        if (imgfile != 0) {
            if (imgfile->isOpen()) {
                // The image was opened successfully.
                currentImage = imgfile;
                imageDevice = imgfile->device();
                markImageSaved();

                loadFileSystem();

                setProgramWindowTitle(imgfile->fileName());
                ui.action_Close_Image->setEnabled(true);
                ui.action_ImageInformation->setEnabled(true);
                ui.action_Save->setEnabled(true);
                ui.actionSave_As->setEnabled(true);
                ui.actionWrite_Disk->setEnabled(true);
                if (imgfile->getFeatures().testFlag(ImageFile::F_Comment))
                    ui.action_ImageInformation->setEnabled(true);
            } else {
                qDebug() << "ERROR: Returned image file is not open.";
            }
        } else {
            qDebug() << "ERROR: Returned image file is NULL.";
        }
    } else {
        qDebug() << "ERROR: Failed to open file. Error code " << QString::number(retval);
        if (retval == errors::E_UNKNOWN_FILE_FORMAT) {
            QMessageBox::warning(this,"Unsupported File Format", "Failed to load disk image: the image file format is unsupported");
        }
    }
}

void Dimg::setProgramWindowTitle() {
    qDebug() << "Update Program Window Title with current filename...";

    if (currentImage)
        setProgramWindowTitle(currentImage->fileName());
    else
        setProgramWindowTitle("");
}

void Dimg::setProgramWindowTitle(QString filename) {
    qDebug() << "Set program window title with filename: " << filename;
    QString appnv = version::application_name + " version " + utils::versionString();
    QString filename_out;

    filename_out = QDir::toNativeSeparators(filename);
    if (isImageDirty())
        filename_out = filename_out + "*";

    if (currentImage != 0) {
        if (currentImage->isOpen()) {
            if (!filename.isEmpty()) {
                setWindowTitle(appnv + " - [" + filename_out + "]");
            }
            else {
                setWindowTitle(appnv + " - [untitled.img*]");
            }
        }
    }
    else {
        setWindowTitle(appnv);
    }
}

void Dimg::loadFSTree(QTreeWidgetItem *parent, QString directory) {
    qDebug() << "Load FS Tree...";
    QList<FSEntry *> entries = fs->getEntryList(directory);
    FSPlugin::Features f = fs->getFeatures();
    qDebug() << "Entries: " << entries.size();
    for (int i = 0; i < entries.size(); i++) {
        FSEntry *e = entries.at(i);

        // Time to build the strings for the entry:
        // Filename - Attributes (rhasd) - size
        QStringList qsl;
        qsl << e->fileName();

        FSPlugin::Attributes a = e->getAttributes();
        QString attributes;

        // Read-Only:
        if (a.testFlag(FSPlugin::A_ReadOnly))
            attributes.append("r");
        else
            attributes.append("-");

        // Hidden:
        if (a.testFlag(FSPlugin::A_Hidden))
            attributes.append("h");
        else
            attributes.append("-");

        // Archive:
        if (a.testFlag(FSPlugin::A_Archive))
            attributes.append("a");
        else
            attributes.append("-");

        // System File:
        if (a.testFlag(FSPlugin::A_System))
            attributes.append("s");
        else
            attributes.append("-");

        // Directory:
        if (a.testFlag(FSPlugin::A_Directory))
            attributes.append("d");
        else
            attributes.append("-");

        qsl << attributes;

        if (!a.testFlag(FSPlugin::A_Directory)) {
            // Compute the size
            quint64 sz = e->getSize();
            QString unit = " B";

            // Are we kilobytes?
            if (sz > 1024) {
                sz /= 1024;
                unit = " KB";
            }

            // Are we megabytes?
            if (sz > 1024) {
                sz /= 1024;
                unit = " MB";
            }

            // are we gigabytes? (thats one huge floppy disk)
            if (sz > 1024) {
                sz /= 1024;
                unit = " GB";
            }

            qsl << QString::number(sz,10) + unit;
        } else { qsl << ""; }

        // Check to see if modification date and time are supported.
        if (f.testFlag(FSPlugin::F_ModDate) &&
            f.testFlag(FSPlugin::F_ModTime)) {
            QDate d = e->getDate(FSEntry::D_ModifiedDate);
            QTime t = e->getTime(FSEntry::D_ModifiedTime);
            QString s = d.toString() + " " + t.toString();
            qsl << s;
        }
        // Otherwise, see if at least creation date and time are
        // supported.
        else if (f.testFlag(FSPlugin::F_CreationDate) &&
                 f.testFlag(FSPlugin::F_CreationTime)) {
            QDate d = e->getDate(FSEntry::D_CreatedDate);
            QTime t = e->getTime(FSEntry::D_CreatedTime);
            QString s = d.toString() + t.toString();
            qsl << s;
        }
        // If creation date and time arent supported either, we will just
        // ignore the date and time
        else {
            qsl << "";
        }


        QTreeWidgetItem *qtwi;
        if (parent == 0) {
            qtwi = new QTreeWidgetItem(ui.treeWidget, qsl,0);
        } else {
            qtwi = new QTreeWidgetItem(parent,qsl,0);
        }

        qtwi->setIcon(0,e->getIcon());
        qtwi->setData(0,Qt::UserRole, e->filePath());

        if (e->isDirectory()) {
            if (e->fileName() != "." && e->fileName() != "..")
                loadFSTree(qtwi,directory + e->fileName() + "/");
        }
    }

    // Now delete all of the FSEntrys:
    while (!entries.isEmpty())
        delete entries.takeFirst();

    ui.treeWidget->resizeColumnToContents(0);
    ui.treeWidget->setColumnWidth(0,ui.treeWidget->columnWidth(0) + 10);
    ui.treeWidget->resizeColumnToContents(1);
    ui.treeWidget->setColumnWidth(1,ui.treeWidget->columnWidth(1) + 10);
    ui.treeWidget->resizeColumnToContents(2);
    ui.treeWidget->setColumnWidth(2,ui.treeWidget->columnWidth(2) + 10);
    ui.treeWidget->resizeColumnToContents(3);
    ui.treeWidget->setColumnWidth(3,ui.treeWidget->columnWidth(3) + 10);
}

void Dimg::resize_columns(const QModelIndex &index) {
    qDebug() << "Resizing columns...";
    ui.treeWidget->resizeColumnToContents(0);
    ui.treeWidget->setColumnWidth(0,ui.treeWidget->columnWidth(0) + 10);
    ui.treeWidget->resizeColumnToContents(1);
    ui.treeWidget->setColumnWidth(1,ui.treeWidget->columnWidth(1) + 10);
    ui.treeWidget->resizeColumnToContents(2);
    ui.treeWidget->setColumnWidth(2,ui.treeWidget->columnWidth(2) + 10);
    ui.treeWidget->resizeColumnToContents(3);
    ui.treeWidget->setColumnWidth(3,ui.treeWidget->columnWidth(3) + 10);
}

QString Dimg::getSelectedFile() {
    qDebug() << "Getting selected file...";
    // Get selected filename - Qt::UserRole

    QTreeWidgetItem *qtwi = ui.treeWidget->currentItem();
    if (qtwi == 0) return "";

    QString d = qtwi->data(0,Qt::UserRole).toString();
    return d;
}

void Dimg::extractFile() {
    qDebug() << "Extracting file...";

    // Figure out just how many files are selected.
    QList<QTreeWidgetItem *> selectedItems = ui.treeWidget->selectedItems();

    // Nothing selected? Nothing to do.
    if (selectedItems.isEmpty()) return;

    // Grab the filesystem entry for the first selected item
    QString fn = selectedItems.first()->data(0, Qt::UserRole).toString();
    if (fn.isEmpty() && selectedItems.count() < 2) return;
    QScopedPointer<FSEntry> firstEntry;
    firstEntry.reset(fs->getEntryByName(fn));

    // One item selected and it isn't a directory?
    if ((selectedItems.count() == 1) && !firstEntry->isDirectory()) {
        // We have a single *FILE* selected. Extract it.
        extractFile(fn);
    } else {
        // We either have multiple files or a directory selected. Either way
        // we use recursive extract.
        treeExtract(selectedItems);
    }
}

void Dimg::treeExtract(QList<QTreeWidgetItem *> listItems) {
    // Extract multiple files. To do this we need to build up a list of
    // exactly what files we need to extract from the disk image.

    QStringList selectedFileList;

    for (int i = 0; i < listItems.count(); i++) {
        QTreeWidgetItem *item = listItems.at(i);

        if (item->childCount() > 0) {
            // If its a directory (has children) then we need to add them
            // to the list of selected items (this is a recursive extract).
            for (int j = 0; j < item->childCount(); j++)
                listItems.append(item->child(j));
        }
        // If its not a directory (doesn't have children)
        selectedFileList << item->data(0,Qt::UserRole).toString();

    }

    extractMultipleFiles(selectedFileList);
}

void Dimg::extractMultipleFiles(QStringList fileNames) {

    // 0. Get the user to select a target directory
    QString lastSaveDir = "";

    lastSaveDir = settings->value("/State/Dirs/LastSaveDir/",
                       settings->value("/State/Dirs/LastOpenDir/","./")).toString();

    qDebug() << "Suggesting directory " << lastSaveDir;

    QString targetDirectory =
            QFileDialog::getExistingDirectory(this,
                                              "Extract files to directory...",
                                              lastSaveDir);
    qDebug() << "User chose to extract to " << targetDirectory;

    if (targetDirectory.isEmpty())
        return;

    // Extraction overwrites files. Check to make sure the user wants to
    // continue if the target directory isn't empty.
    QDir targetDir(targetDirectory);
    if (!targetDir.entryList().isEmpty()) {
        int result = QMessageBox::warning(this, "Not Empty",
                                          "The selected directory is not empty."
                                          " Extracting files here will "
                                          "overwrite any existing files with "
                                          "the same name. Do you wish to "
                                          "continue?",
                                          QMessageBox::Ok | QMessageBox::Cancel);
        if (result == QMessageBox::Cancel)
            return;
    }

    // 1. Convert all the names to directory entries (with attributes and stuff)
    QList<FSEntry *> entries;

    foreach(QString item, fileNames) {
        entries.append(fs->getEntryByName(item));
    }

    // 2. Separate out the directories and the files
    QList<FSEntry *> files;
    QList<FSEntry *> directories;

    foreach(FSEntry *ent, entries) {
        if (ent->isDirectory())
            directories.append(ent);
        else
            files.append(ent);
    }

    // 3. Create the directories
    foreach(FSEntry *ent, directories) {
        QString cleanPath = QDir::toNativeSeparators(ent->filePath());
        cleanPath = QDir::cleanPath(targetDirectory + cleanPath);

        qDebug() << "mkdir" << cleanPath;
        QDir dir;
        bool result = dir.mkpath(cleanPath);
        if (!result) qWarning() << "Failed to create directory.";
    }

    // 4. Extract the files
    foreach(FSEntry *ent, files) {
        QString localName = QDir::toNativeSeparators(ent->filePath());
        localName = QDir::cleanPath(targetDirectory + QDir::separator() + localName);

        qDebug() << "extract" << ent->filePath() << "to" << localName;
        fs->copyFile(ent->filePath(), localName);
    }


    // 5. Now we clean up so as to not leak memory
    while (!entries.isEmpty())
        delete entries.takeFirst();

    // 6. And update settings
    if (settings->contains("/State/Dirs/LastSaveDir/"))
        settings->setValue("/State/Dirs/LastSaveDir/",
                           QFileInfo(targetDirectory).absolutePath());
    else
        settings->setValue("/State/Dirs/LastOpenDir/",
                           QFileInfo(targetDirectory).absolutePath());
}

void Dimg::showFileProperties() {
    qDebug() << "Showing File Properties...";
    QString fn = getSelectedFile();
    if (fn.isEmpty()) return;
    showFileProperties(fn);
}

void Dimg::extractFile(QString filename) {
    qDebug() << "Extracting file " << filename;

    QString lastSaveDir = "";

    lastSaveDir = settings->value("/State/Dirs/LastSaveDir/",
                       settings->value("/State/Dirs/LastOpenDir/","./")).toString();

    lastSaveDir = lastSaveDir + QDir::separator() + QFileInfo(filename).fileName();

    qDebug() << "Suggesting filename " << lastSaveDir;

    QString dest_fn = QFileDialog::getSaveFileName(this,"Extract File...",lastSaveDir);

    qDebug() << "User chose to save as " << dest_fn;

    if (dest_fn.isEmpty()) return;

    fs->copyFile(filename,dest_fn);

    if (settings->contains("/State/Dirs/LastSaveDir/"))
        settings->setValue("/State/Dirs/LastSaveDir/",QFileInfo(dest_fn).absolutePath());
    else
        settings->setValue("/State/Dirs/LastOpenDir/",QFileInfo(dest_fn).absolutePath());
}

void Dimg::showFileProperties(QString filename) {
    qDebug() << "Showing file properties for " << filename;
    FileProperties *fp = new FileProperties(this);
    FSEntry2 *e = fs->getEntryByName(filename);
    if (e == 0) return;
    fp->setFileEntry(e);
    fp->show();
}

void Dimg::closeEvent(QCloseEvent *event) {
    qDebug() << "Close event: application exiting.";

    // If there is an unsaved image open, ask if it should be saved.
    if (prompt_image_save() == errors::E_USER_CANCELED) {
        if (event)
            event->ignore();
        return;
    }

    // We can't let closeImage do the saving as we need to know what option
    // the user chooses so we can cancel the close event if need be.
    closeImage(false);

    // Perform exit stuff for the filesystem plugin
    if (fs) {
        fs->exiting();
        delete fs;
        fs = 0;
    }

    if (currentImage) {
        delete currentImage;
        currentImage = 0;
    }

    // Exit
    pApp->exit(0);
}

void Dimg::closeImage(bool save) {
    qDebug() << "Close image requested";

    /*
     * If image is unsaved, run saveAs.
     */
    if (save && currentImage != 0) {
        if (imageUnsaved) {
            if (prompt_image_save() != errors::E_SUCCESS) return;
        } else if (isImageDirty()) {
            if (prompt_image_save() != errors::E_SUCCESS) return;
        }
    }

    if (fs) {
        fs->close();
        ui.treeWidget->clear();
    }

    // the imageDevice belongs to the ImageFile so don't delete it or anything.
    if (imageDevice) imageDevice = 0;

    if (currentImage) {
        currentImage->close();
        delete currentImage;
        currentImage = 0;
        markImageSaved();
    }

    ui.action_Close_Image->setEnabled(false);
    ui.action_ImageInformation->setEnabled(false);
    ui.action_Save->setEnabled(false);
    ui.actionSave_As->setEnabled(false);
    ui.actionWrite_Disk->setEnabled(false);
    ui.action_ImageInformation->setEnabled(false);
    setFSEnabled(false);
    volumeLabel->setText("");
    setProgramWindowTitle();
}

void Dimg::createImage(){
    qDebug() << "Creating blank image...";

    currentImage = new CacheImageFile(this);
    imageDevice = currentImage->device();
    markImageUnsaved();

    ui.action_Close_Image->setEnabled(true);
    ui.action_ImageInformation->setEnabled(true);
    ui.action_Save->setEnabled(true);
    ui.actionSave_As->setEnabled(true);
    ui.actionWrite_Disk->setEnabled(true);
    setProgramWindowTitle();
}

void Dimg::loadFileSystem(){
    qDebug() << "Loading filesystem...";
    if (fs)
        if (!fs->isClosed()) {
            qDebug() << "Closing previous filesystem instance...";
            fs->close();
        }

    qDebug() << "Loading filesystem...";
    if (fs) {
        if (imageDevice) {
            qDebug() << "Loading filesystem";
            fs->load(imageDevice);
        }
    }

    if (fs->isClosed()) {
        // Filesystem failed to open.
        setFSEnabled(false);
        statusBar()->showMessage("Filesystem failed to load.");
        return;
    }

    setFSEnabled(true);

    qDebug() << "Displaying File Tree...";
    ui.treeWidget->clear();
    if (fs) {
        loadFSTree(0,"/");
        ui.actionExtract_File->setEnabled(true);
        ui.actionFile_Properties->setEnabled(true);
    }

    FSPlugin::Features f = fs->getFeatures();

    if (f.testFlag(FSPlugin::F_VolumeLabel)) {
        qDebug() << "Displaying Volume Label...";
        volumeLabel->setText("Volume Label: " + fs->volumeLabel());
    }
}

void Dimg::setFSEnabled(bool enabled) {
    volumeLabel->setVisible(enabled);
    ui.treeWidget->setEnabled(enabled);
    ui.actionExtract_File->setEnabled(enabled);
    ui.actionFile_Properties->setEnabled(enabled);
}

void Dimg::showAbout() {
    qDebug() << "Showing about screen...";
    QString fsPluginName = "None Loaded", fsPluginVersion;
    QString dqPluginName = "None Loaded", dqPluginVersion;


    AboutDialog ad;
    if (fs) {
        fsPluginName = fs->getPluginName();
        fsPluginVersion = fs->getVersionString();
    }
    if (imgr->driver_ready()) {
        dqPluginName = imgr->get_driver_name();//dqdrv->getPluginName();
        dqPluginVersion = imgr->get_driver_version();//dqdrv->getVersionString();
    }

    ad.setData(fsPluginName, fsPluginVersion, dqPluginName, dqPluginVersion);
    ad.exec();
}

int Dimg::saveCurrentImage() {
    if (imageUnsaved) {
        // No filename is set its never been saved before. Force Save-As.
        return saveCurrentImageAs();
    }

    int retval = currentImage->save();

    if (retval != errors::E_SUCCESS) return retval;

    setProgramWindowTitle();    // Update the window title
    return retval;   // Return success
}

int Dimg::saveCurrentImageAs() {

    ImageFile *newImageFile = 0;
    int retval = ifu.saveImageAs(&currentImage, &newImageFile);

    if (retval != errors::E_SUCCESS ||
        newImageFile == 0)
        return retval;

    if (!newImageFile->isOpen()) return retval;

    // Ok. We seem to have a new image file. Lets use that instead.
    currentImage->close();
    delete currentImage;
    currentImage = newImageFile;
    imageDevice = currentImage->device();
    markImageSaved();

    if (currentImage->getFeatures().testFlag(ImageFile::F_Comment))
        ui.action_ImageInformation->setEnabled(true);

    setProgramWindowTitle();

    return errors::E_SUCCESS;
}

int Dimg::prompt_image_save() {

    if (currentImage == 0) return errors::E_NO_FILE_OPEN;

    if (!currentImage->isOpen()) return errors::E_NO_FILE_OPEN;

    // Is there an image open? Is it dirty? If so, save it.
    if (isImageDirty() || imageUnsaved) {
        qDebug() << "Unsaved image open - prompting to save...";
        QMessageBox msgbox;
        msgbox.setText("The current image has been modified. Do you want to save before closing it?");
        msgbox.setStandardButtons(QMessageBox::Save | QMessageBox::Discard | QMessageBox::Cancel);
        msgbox.setDefaultButton(QMessageBox::Save);
        msgbox.setIcon(QMessageBox::Question);
        int ret = msgbox.exec();

        switch (ret) {
            case (QMessageBox::Save):
                qDebug() << "-> User chose to save.";
                return saveCurrentImage();
                break;
            case (QMessageBox::Cancel):
                qDebug() << "-> User chose to cancel.";
                return errors::E_USER_CANCELED;
                break;
            case (QMessageBox::Discard):
                qDebug() << "-> User chose to discard.";
                break;
            default:
                break;
        }
    }
    return errors::E_SUCCESS;
}

int Dimg::promptImageSaveAs() {

    if (currentImage == 0) return errors::E_NO_FILE_OPEN;

    if (!currentImage->isOpen()) return errors::E_NO_FILE_OPEN;

    // Is there an image open? Is it dirty? If so, save it.
    if (isImageDirty() || imageUnsaved) {
        qDebug() << "Unsaved image open - prompting to save...";
        QMessageBox msgbox;
        msgbox.setText("The current image has been modified. Do you want to save before closing it?");
        msgbox.setStandardButtons(QMessageBox::Save | QMessageBox::Discard | QMessageBox::Cancel);
        msgbox.setDefaultButton(QMessageBox::Save);
        msgbox.setIcon(QMessageBox::Question);
        int ret = msgbox.exec();

        switch (ret) {
            case (QMessageBox::Save):
                qDebug() << "-> User chose to save.";
                return saveCurrentImageAs();
                break;
            case (QMessageBox::Cancel):
                qDebug() << "-> User chose to cancel.";
                return errors::E_USER_CANCELED;
                break;
            case (QMessageBox::Discard):
                qDebug() << "-> User chose to discard.";
                break;
            default:
                break;
        }
    }
    return errors::E_SUCCESS;
}

void Dimg::imageCurrentDrive(){
    qDebug() << "Imaging current drive...";
    /* Ok, here is what we need to do:
        1. If there is an image open, close it.
        1.1. If the image is dirty, ask to save it.
        1.2. Clear buffers, close filesystem, etc.
        2. Prepare a QBuffer to image into
        3. Tell The Imager to Get Imaging!
     */

    // First, we cant image if the Imagers driver isn't ready.
    if (!imgr->driver_ready()) {
        qDebug() << "CRITICAL: Unable to image - Imager Driver not ready.";
        QMessageBox::critical(0,"Unable to Create Image", "No drive access plugin is currently active. Without a drive access plugin\n" + version::application_name + " can not read or write images to drives.");
        return;
    }

    // Ask if we should save the image
    if (prompt_image_save() == errors::E_USER_CANCELED)
        return;

    // Ok, the current image has been dealt with. Close it.
    closeImage();
    qDebug() << "Image Closed.";

    // Now we need to prepare something to image the disk into
    createImage();

    // And lastly, we need to actually image the disk in the current drive.
    int errcode = imgr->read_image(imageDevice);
    imageDevice->seek(0);

    qDebug() << "Read Image completed with code " << errcode;

    if (errcode) {
        if (errcode == errors::E_NO_DRIVE_SELECTED) {
            QMessageBox::warning(this,"No Drive Selected",
                                 "No drive is currently selected. Select a "
                                 "drive before trying to perform an operation "
                                 "on it.");
        }

        // We do not display an error here as The Imager should have already
        // done it. Just close the image and move on.
        closeImage(false);
    } else {
        qDebug() << "Imaging Successful";

        // Load the Filesystem
        loadFileSystem();

        // done!
        QMessageBox::information(0,"Complete.", "Disk Imaging Complete.");
    }
}

void Dimg::writeCurrentDrive() {
    qDebug() << "Writing current drive...";

    if (imageDevice == 0) {
        qWarning() << "Image device is NULL. Aborting.";
        return;
    }

    int errcode = imgr->write_image(imageDevice);
    imageDevice->seek(0);
    if (!errcode) {
        qDebug() << "Imaging Successful";

        // done!
        QMessageBox::information(0,"Complete.", "Disk Imaging Complete.");
    } else {
        if (errcode == errors::E_NO_DRIVE_SELECTED) {
            QMessageBox::warning(this,"No Drive Selected",
                                 "No drive is currently selected. Select a "
                                 "drive before trying to perform an operation "
                                 "on it.");
        }
    }
}

void Dimg::showImageInformation() {
    ImageInformation ii(currentImage,this);

    int ret = ii.exec();

    if (ret == QDialog::Accepted) {
        if (currentImage->getFeatures().testFlag(ImageFile::F_Comment)) {
            QString currentComment = currentImage->getComment();
            QString newComment = ii.getComment();
            if (currentComment != newComment) {
                currentImage->setComment(newComment);
            }
        }
    }

    setProgramWindowTitle();
}

void Dimg::showSettings() {
    SettingsDialog sd;
    sd.exec();
}

bool Dimg::isImageDirty() {
    if (currentImage != 0)
        return currentImage->isDirty();
    return false;
}
