/*
-----------------------------------------------------
   Application Name: "DIMG"
Application Version: 1.1.0
          File Name: imagefilefactory.h
         Start Date: 25/02/2010
       Last Changed: $LastChangedDate: 2011-04-09 00:17:11 +1200 (Sat, 09 Apr 2011) $
           Revision: $Revision: 295 $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2008, 2009, 2010
            License: GNU General Public License
   File Description: ImageFileFactory Class header
-----------------------------------------------------
        This file is part of "DIMG".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

        File Notes:
        ------------------
*/

#ifndef IMAGEFILEFACTORY_H
#define IMAGEFILEFACTORY_H

#include <QObject>
#include <QString>
#include <QIODevice>
#include <QPointer>
#include <QStringList>
#include <QDir>
#include <QPluginLoader>
#include <QtDebug>

#include <imagefile.h>
#include <imagefileplugin.h>

#include "errors.h"

/**
\brief This class handles loading image files using format translators.

The ImageFileFactory selects an appropriate translator based on the extension
of the filename provided to it. It then passes it to the calling function as
a QIODevice.
 */
class ImageFileFactory : public QObject
{
Q_OBJECT
public:

    /** File Save hints
     */
    enum FileHint {
        F_NoHints      = 0x00000000, /*!< No hints */
        F_DontCache    = 0x00000001  /*!< Don't cache the image file in memory */
    };
    Q_DECLARE_FLAGS(FileHints, FileHint)

    /**
      Constructs a new ImageFileFactory.
    \param parent The parent object
     */
    explicit ImageFileFactory(QObject *parent = 0);

    /** Destroys the ImageFileFactory.
     */
    ~ImageFileFactory();

    /** Loads the specified image file. This function will guess which
      format translator to use based on filename extension. The translator
      is then passed out as an ImageFile.

        \param filename The file to open.
      \param imageFile The ImageFile which can be used to gain read and
      write the image file is passed out here.
        \returns An error code.
     */
    int loadImage(QString filename, ImageFile** imageFile);

    /** Creates a new image file. The format used will be based on the file
      extension. The translator is passed out as an ImageFile.

      \param filename The file to create.
      \param imageFile The ImageFile which can be used to gain read and
      write the image file is passed out here.
      \param expectedSize The expected size
      \returns An error code.
     */
    int createImage(QString filename,
                    ImageFile** imageFile,
                    long long expectedSize);

    /** Creates a new image file. The format used will be based on the file
      extension. The translator is passed out as an ImageFile.

      \param filename The file to create.
      \param imageFile The ImageFile which can be used to gain read and
      write the image file is passed out here.
      \param hints File hints (such as dont cache in memory, etc).
      \returns An error code.
     */
    int createImage(QString filename,
                    ImageFile** imageFile,
                    long long expectedSize,
                    FileHints hints);

    /** Returns a combined list of filename filters from all loaded image
      format plugins. These are suitable for use in file dialogs, etc.
      \param includeAllFilter If true, an "All Supported Images" filter will be included.
      \returns A list of filename filters
     */
    QStringList getFileNameFilters(bool includeAllFilter = true);

    /** Loads all image file format plugins.
      \param pluginDirectory The directory to scan for plugins.
      \returns An error code.
     */
    int loadPlugins(QString pluginDirectory);
signals:

public slots:

private:

    QString getExtensionsFromFormatString(QString formatString);

    QList<ImageFilePlugin *> plugins;
    ImageFilePlugin *rawPlugin;
};

#endif // IMAGEFILEFACTORY_H
