/*
-----------------------------------------------------
   Application Name: "DIMG"
         Start Date: 09/07/2010
       Last Changed: $LastChangedDate: $
           Revision: $Revision: $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2010
            License: GNU General Public License
-----------------------------------------------------
        This file is part of "DIMG".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

        File Notes:
        ------------------
*/

#ifndef BIOSPARAMBLOCK_H
#define BIOSPARAMBLOCK_H

#include <QObject>
#include <QIODevice>
#include <QPointer>
#include <QByteArray>

#include "errors.h"

using namespace FilesystemFAT::Errors;

namespace FilesystemFAT {

    /** This class reads the BIOS Parameter Block as used by FAT filesystems. It is
      currently targeted towards FAT12 but does contain some (untested) code for
      FAT16/FAT32 too. Currently this implementation assumes an unpartitioned
      filesystem image and so probably only works with floppy disks.

      This implementation is not compatible with the BIOS Parameter Block used by
      Apricot PCs or anything else unusual.

      Accessor methods starting with ext are for fields in the Extended BIOS Parameter
      Block which this class also reads. There are two variants (three depending on
      how you count it) of the Extended BIOS Parameter Block - one (or two) used by
      FAT12/16 and one used by FAT32. They are different and entirely incompatible
      on-disk structures. The FAT32 one contains most of the fields in the FAT12/16
      one (at different locations) but the same is not true the other way round.

      The final three fields of the FAT12/16 EBPB (ID, Volume Label, FAT Filesystem
      type) are not always present.

      Functions starting with ext12 are for fields at FAT12/16 EBPB locations where
      as ext32 are for fields at the FAT32 locations. The FAT32 ones are generally
      not implemented at this stage.

      In the future some form of autodetection may be implemented in which case
      functions for common fields will be merged.
     */
    class BiosParamBlock : public QObject
    {
    Q_OBJECT
    public:
        /** Constructs a new BiosParamBlock object.

          \param parent The parent QObject.
         */
        explicit BiosParamBlock(QObject *parent = 0);

        /** Loads the BIOS Parameter Block from the specified QIODevice.

          \param dev The QIODevice to read the BIOS Parameter Block from.
          \param offset Where the BIOS Parameter Block starts in the QIODevice.
          \returns An error code.
         */
        ReturnCode load(QIODevice *dev, int offset = 0);

        /** Closes the currently open BIOS Parameter Block (if any).
         */
        void close();

        /** Returns true if the BIOS Parameter Block is currently open.
          \returns True if the BIOS Parameter Block is currently open, otherwise False.
         */
        bool isOpen() const {return open;}

        /** Reloads the cached values from the QIODevice supplied to load().

          \returns An error code.
         */
        ReturnCode reloadCache();

        /** Returns the last error to occur.
          \returns The current error code.
         */
        ReturnCode lastErr() {return lastReturnCode;}

        /** Reads the OEM_ID field. MS-DOS checks this field to figure out
          which other parts of the BIOS Parameter Block can be relied on.
          Common values are "IBM  3.3", "MSDOS5.0", "MSWIN4.1" and "mkdosfs".

          Its offset is 0x03 (there is a three-byte jump instruction at 0x00)
          and is 8 bytes long. It isn't really part of the BIOS Parameter
          Block.

          \returns The value in the OEM ID field.
         */
        QString oemID() const;

        /** Reads the Bytes Per Sector field. This is really the first value
          of the BIOS Parameter Block. It starts at 0x0B and is two bytes
          long.

          512 is a common value.

          \returns The number of bytes per sector.
         */
        unsigned int bytesPerSector() const;

        /** Reads the Sectors Per Cluster field. This field allows powers of
          two from 1 to 128. The value must not be such that the number of
          bytes per cluster is greater than 32KB.

          This field starts at 0x0D and is one byte long.

          \returns The number of sectors per cluster.
         */
        unsigned int sectorsPerCluster() const;

        /** Reads the number of Reserved Sectors. This is the number of
          sectors before the first FAT in the file system image. Normally
          1 for FAT12/FAT16 and 32 for FAT32.

          This field starts at 0x0E and runs for two bytes.
          \returns The number of reserved sectors.
         */
        unsigned int reservedSectors() const;

        /** Reads the number of File Allocation Tables. This is almost
          always two.

          This field starts at 0x10 and is one byte long.
          \returns The number of File Allocation Tables.
         */
        unsigned int numberOfFATs() const;

        /** Reads the maximum number of root directory entries. This is only used by
          FAT12 and FAT16, where the root directory is handled specially. It should
          always be 0 for FAT32 filesystem images.

          This value should be such that the root directory always ends on a sector
          boundary. That is, the root directory size should be a multiple of the
          sector size.

          224 is typical for floppy disks.

          \returns The maximum number of root directory entries.
         */
        unsigned int maxRootEntries() const;

        /** Total number of sectors.

          This is read from one of two locations in the BIOS Parameter Block. The
          two byte value starting at 0x13 will be checked first. If this is zero
          then it is assumed the total sector count is greater than 65535 so the
          four byte value starting at 0x20 will be returned instead.
         */
        unsigned long totalSectors() const;

        /** The media descriptor.
           <table>
             <tr>
               <td>0xF0</td>
               <td>3.5" double sided, 80 tracks per side, 18 or 36 sectors per track
                   (1.44MB or 2.88MB). 5.25" double sided, 80 tracks per side, 15
                   sectors per track (1.2MB). Also used for other media types</td>
             </tr>
             <tr>
               <td>0xF8</td>
               <td>Fixed disk</td>
             </tr>
             <tr>
               <td>0xF9</td>
               <td>3.5" Double sided, 80 tracks per side, 9 sectors per track (720K).
                   5.25" Double Sided, 80 tracks per side, 15 sectors per track (1.2MB).
                   </td>
             </tr>
             <tr>
               <td>0xFA</td>
               <td>5.25" Single sided, 80 tracks per side, 8 sectors per track (320K)
                   </td>
             </tr>
             <tr>
               <td>0xFB</td>
               <td>3.5" Double sided, 80 tracks per side, 8 sectors per track (640K)
                   </td>
             </tr>
             <tr>
               <td>0xFC</td>
               <td>5.25" Single Sided, 40 tracks per side, 9 sectors per track (180K)
                   </td>
             </tr>
             <tr>
               <td>0xFD</td>
               <td>5.25" Double Sided, 40 tracks per side, 9 sectors per track (360K).
                   Also used for 8"</td>
             </tr>
             <tr>
               <td>0xFE</td>
               <td>5.25" Single sided, 40 tracks per side, 8 sectors per track (160K).
                   Also used for 8"</td>
             </tr>
             <tr>
               <td>0xFF</td>
               <td>5.25" Double sided, 40 tracks per side, 8 sectors per track (320K).
                   </td>
             </tr>
           </table>
           This value should also be repeated as the first byte of each copy of the
           File Allocation Table. Some operating systems such as MSX-DOS 1.0 ignore
           this field and pull the value from the first byte of the FAT to figure
           out the file system details.

           This field starts at 0x15 and is a single byte long.

           \returns The media descriptor.
         */
        unsigned char mediaDescriptor() const;

        /** Number of sectors per File Allocation Table for FAT12/FAT16.

          This field starts at 0x16 and is two bytes long.

          \returns The number of sectors per FAT.
         */
        unsigned int sectorsPerFAT() const;

        /** The number of sectors per track. This value is only useful on disks
          with geometry.

          This field starts at 0x18 and is two bytes long.

          \returns The number of sectors per track.
         */
        unsigned int sectorsPerTrack() const;

        /** Number of heads. This is only useful on disks with geometry.

          This field starts at 0x1A and is two bytes long.

          \returns The number of heads.
         */
        unsigned int numberOfHeads() const;

        /** Count of hidden sectors preceding the partition that contains this
          FAT volume. This field should always be zero on non-partitioned media.

          This field starts at 0x1C and is four bytes long.

          \returns The number of hidden sectors.
         */
        unsigned long hiddenSectors() const;

        /** The physical drive number in the Extended BIOS Parameter Block. This field
          is present in both FAT12/16 and FAT32 versions. Values are 0x00 for removable
          media and 0x80 for hard disks.

          This field is located at either 0x24 (FAT12/FAT16) or 0x40 (FAT32) and is
          one byte long.

          \returns The physical drive number.
         */
        unsigned char extPhysicalDriveNumber() const;

        /** The reserved field in the FAT12/16 Extended BIOS Parameter Block.
          Bits 0 and 1 used as dirty flags and request a disk check and surface
          scan respectively on boot. See ext12ChkdskRequested() and
          ext12SurfaceScanRequested() to pull out these values.

          It is located at 0x25 and is one byte long.

          \returns The contents of the FAT12/16 Extended BIOS Parameter Block
          reserved field.
         */
        unsigned char ext12Reserved() const;

        /** Chkdsk requested flag. This is stored as bit 0 in the Reserved field
          of the FAT12/16 Extended BIOS Parameter Block by Windows NT to request
          that chkdsk be run on startup. It shows that the volume was not cleanly
          unmounted.

          \returns True if bit 0 in the FAT12/16 Extended BPB reserved field is set.
         */
        bool ext12ChkdskRequested() const;

        /** Surface Scan requested flag. This is stored as bit 1 in the Reserved
          field of the FAT12/16 Extended BIOS Parameter Block by Windows NT to
          request that a surface scan be run on startup.

          \returns True if bit 1 in the FAT12/16 Extended BPB reserved field is set.
         */
        bool ext12SurfaceScanRequested() const;

        /** The extended boot signature field. Should be 0x29 indicating that
          the ID/Serial Number, Volume Label and FAT Filesystem type fields are
          present.

          This is only present in the FAT12/FAT16 Extended BIOS Parameter Block.

          \returns The value in the Extended Boot Signature field.
         */
        unsigned char ext12ExtendedBootSignature() const;

        /** The Volume ID field. This should be present if ext12ExtendedBootSignature()
          gives a value of 0x29 but there is really no guarantee. It is often called
          the Volume Serial Number.

          It lives at 0x26 and is four bytes long.

          \returns The contents of the Volume ID field. This is value is probably invalid
          if ext12ExtendedBootSignature() returns a value other than 0x29.
         */
        unsigned int ext12VolumeID() const;

        /** The Volume Label. This is padded wth blanks (0x20) up to 11 characters. It
          starts at 0x2B and is 11 bytes long.

          This value may not be entirely trustworthy and is not always present. If the
          root directory contains a file with the Volume ID attribute set its probably
          better to use that than this value.

          This field may not exist if ext12ExtendedBootSignature() returns a value other
          than 0x29 (but even if 0x29 does come back there is probably still no guarantee
          that what ever formatted the disk added this field).

          \returns The Volume Label. This value is probably garbage if ext12ExtendedBootSignature()
          returns a value other than 0x29.
         */
        QString ext12VolumeLabel() const;

        /** This is the FAT Filesystem Type field. It is padded with blanks (0x20) up to eight
          characters. It lives at 0x36 and is eight bytes long.

          Common values are "FAT12   " and "FAT16   ". It should
          not be used to determine the drive type but some utilities do use it this way.
         */
        QString ext12FSType() const;

    private:
        bool open;  /*!< If the BPB is open or not. Set by load() and unset by close() */
        QPointer<QIODevice> ioDevice;   /*!< The QIODevice we will read from where necessary */
        unsigned int offset; /*!< Where the BIOS Parameter Block starts */

        /** This is the current error code.
         */
        ReturnCode lastReturnCode;

        QString cOEMID; /*!< Cached OEM ID value */
        unsigned int cBytesPerSector;   /*!< Cached Bytes Per Sector value */
        unsigned int cSectorsPerCluster; /*!< Cached Sectors Per Cluster value */
        unsigned int cReservedSectors; /*!< Cached Reserved Sectors value */
        unsigned int cNumberOfFATs; /*!< Cached Number of FATs value */
        unsigned int cMaxRootEntries; /*!< Cached Max Root Entries value */
        unsigned long cTotalSectors; /*!< Cached Total Sectors value */
        unsigned char cMediaDescriptor; /*!< Cached Media Descriptor value */
        unsigned int cSectorsPerFAT; /*!< Cached Sectors Per FAT value */
        unsigned int cSectorsPerTrack; /*!< Cached Sectors Per Track value */
        unsigned int cNumberOfHeads;    /*!< Cached Number of Heads value */
        unsigned long cHiddenSectors; /*!< Cached number of hidden sectors value */
        unsigned char cExt12PhysicalDriveNumber; /*!< Cached EBPB-12 Physical Drive Number */
        unsigned char cExt12Reserved; /*!< Cached EBPB-12 Reserved value */
        bool cExt12ChkdskRequested; /*!< Cached EBPB-12 ChkDsk Requested flag */
        bool cExt12SurfaceScanRequested; /*!< Cached EBPB-12 Surface Scan Requested flag */
        unsigned char cExt12ExtendedBootSignature; /*!< Cached EBPB-12 Extended Boot Sig value */
        unsigned int cExt12VolumeID; /*!< Cached EBPB-12 Volume ID */
        QString cExt12VolumeLabel; /*!< Cached EBPB-12 Volume Label */
        QString cExt12FSType; /*!< Cached EBPB-12 Filesystem Type value */
    };
}

#endif // BIOSPARAMBLOCK_H
