/*
-----------------------------------------------------
   Application Name: "DIMG"
         Start Date: 09/07/2010
       Last Changed: $LastChangedDate: $
           Revision: $Revision: $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2010
            License: GNU General Public License
-----------------------------------------------------
        This file is part of "DIMG".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

        File Notes:
        ------------------
*/

#include "fatfs.h"
#include "fat12.h"

#include <QStringList>
#include <QtDebug>


using namespace FilesystemFAT;

FatFS::FatFS(QObject *parent) :
        QObject(parent)
{
    bpb = NULL;
    da = NULL;
    dev = NULL;
    root = NULL;
    fs_open = false;
}

bool FatFS::open() {

    // Can't open if it is already open.
    if (isOpen()) {
        qWarning() << "Filesystem is already open.";
        lastErr = Errors::FS_IS_OPEN;
        return false;
    }

    // Can't open if we don't have something to open it from
    if (dev == NULL) {
        qWarning() << "No device specified.";
        lastErr = Errors::NO_DEVICE;
        return false;
    }

    if (!dev->isOpen()) {
        qWarning() << "The supplied device is closed.";
        lastErr = Errors::DEV_CLOSED;
        return false;
    }

    // Everything seems OK. Proceed with open.
    qDebug() << "Loading filesystem...";

    // BIOS Parameter Block (Volume Boot Record)
    qDebug() << "Loading BIOS Parameter Block...";
    bpb = new BiosParamBlock(this);
    Errors::ReturnCode ret = bpb->load(dev);
    if (ret == Errors::BPB_WARNING) {
        // BPB_WARNINGSs are potentially recoverable errors encountered in the
        // BIOS Parameter Block. Its stuff like if the root directory size is a
        // multiple of the sector size. Nothing too major.
        qWarning() << "An error was detected reading the BPB. This may cause "
                "problems further down the line. Attempting to proceed...";
    } else if (ret != Errors::SUCCESS) {
        qWarning() << "Failed to open BIOS Parameter Block. Aborting.";
        delete bpb;
        lastErr = ret;
        return false;
    }

    // File Allocation Table
    qDebug() << "Loading Primary File Allocation Table...";
    AbstractFAT * fat = new Fat12(dev,bpb,0,this);
    ret = fat->lastError();
    if (ret != Errors::SUCCESS) {
        qWarning() << "Failed to open File Allocation Table. Aborting.";
        delete fat;
        delete bpb;
        lastErr = ret;
        return false;
    } else {
        fats.append(fat);
    }

    // Load any additional FATs. These aren't quite as important as the first
    // one so we wont give up if one fails to load.
    if (bpb->numberOfFATs() > 1) {
        for (unsigned int i = 1; i < bpb->numberOfFATs(); i++) {
            qDebug() << "Loading File Allocation Table" << i+1;
            fat = new Fat12(dev,bpb,i,this);
            ret = fat->lastError();

            // If it loaded successfully, store it away. Otherwise delete it.
            if (ret == Errors::SUCCESS)
                fats.append(fat);
            else {
                qWarning() << "Failed to load File Allocation Table" << i+1;
                delete fat;
            }
        }
    }


    // Data Area (reads clusters)
    qDebug() << "Loading Data Area...";
    da = new DataArea(dev,bpb,this);
    ret = da->lastError();
    if (ret != Errors::SUCCESS) {
        qWarning() << "Failed to open Data Area. Aborting.";
        delete da;
        delete fat;
        delete bpb;
        lastErr = ret;
        return false;
    }

    // Root directory (so we can create other directories)
    qDebug() << "Loading Root Directory...";
    root = new Directory(this,this);
    ret = root->lastError();
    if (ret != Errors::SUCCESS) {
        qWarning() << "Failed to open the Root Directory. Aborting.";
        delete root;
        delete da;
        delete fat;
        delete bpb;
        lastErr = ret;
        return false;
    }

    // Finished.
    qDebug() << "FatFS initialised successfully.";
    lastErr = Errors::SUCCESS;
    fs_open = true;
    return true;
}

bool FatFS::isOpen() {
    return fs_open;
}

Directory* FatFS::createDirectory(QString directoryName) {
    qDebug() << "----> FatFS::createDirectory(" << directoryName << ")";
    qDebug() << "Searching filesystem for directory " << directoryName;

    if (!isOpen()) {
        qWarning() << "Filesystem is not open.";
        lastErr = Errors::FS_IS_CLOSED;
        return NULL;
    }

    // If the root directory was requested, create one.
    if (directoryName == "/") {
        lastErr = Errors::SUCCESS;
        return new Directory(this,this);
    }

    // This is the path we must follow
    QStringList path = directoryName.split('/');

    // Clear out blank entries.
    for (int i = 0; i < path.count(); i++)
        if (path.at(i).isEmpty())
            path.removeAt(i);

    if (path.length() < 1) {
        lastErr = Errors::FILE_NOT_FOUND;
        return NULL;
    }

    // We'll start right here.
    Directory * prevdir = NULL;
    Directory * curdir = root;

    // start searching
    while (!path.isEmpty()) {
        // Grab the next directory name
        QString nextDirectory = path.takeFirst();

        // An empty directory name is probably just a typo - continue.
        if (nextDirectory.isEmpty()) continue;

        prevdir = curdir;
        curdir = prevdir->getSubdir(nextDirectory);
        if (curdir == NULL) {
            /* getSubdir returned null - the subdirectory couldnt be
               found. That means that the directory we are searching for must
               not exist. Return null to indicate this.
             */
            lastErr = Errors::FILE_NOT_FOUND;
            if (prevdir != NULL && prevdir != root) delete prevdir;

            return NULL;
        }

        if (!path.isEmpty() && prevdir != root) {
            /* If we aren't currently looking at the last directory then
             * we don't care about it. Throw it away.
             */

            delete prevdir;
        }
    }

    lastErr = Errors::SUCCESS;
    return curdir;
}

FsFile *FatFS::createFile(QString fileName) {
    QStringList parts = fileName.split("/");
    QString name = parts.takeLast();
    QString parentDir = parts.join("/");
    if (parentDir.isEmpty()) parentDir = "/";

    QScopedPointer<Directory> dir;
    dir.reset(createDirectory(parentDir));
    if (lastError() != Errors::SUCCESS) return NULL;

    FsFile *file = dir->getFile(name);
    if (dir->lastError() != Errors::SUCCESS) {
        lastErr = dir->lastError();
        return NULL;
    } else {
        return file;
    }

}

AbstractFAT* FatFS::getFAT(int number) {
    return fats.at(number);
}

BiosParamBlock* FatFS::getBPB() {
    return bpb;
}

DataArea* FatFS::getDataArea() {
    return da;
}

void FatFS::setDevice(QIODevice *device) {
    if (isOpen()) {
        qWarning() << "Unable to set device while Filesystem is open.";
        lastErr = Errors::FS_IS_OPEN;
        return;
    }
    dev = device;
}

QIODevice* FatFS::getDevice() {
    return dev;
}

QString FatFS::getVolumeLabel() {
    if (!isOpen()) {
        lastErr = Errors::FS_IS_CLOSED;
        return "";
    }
    lastErr = Errors::SUCCESS;

    // If the BIOS Parameter Block doesn't have a volume label then get it
    // from the root directory instead.
    if (!bpb->ext12VolumeLabel().isEmpty())
        return bpb->ext12VolumeLabel();
    else
        return root->getVolumeLabel();
}
