/*
-----------------------------------------------------
   Application Name: "DIMG"
         Start Date: 09/07/2010
       Last Changed: $LastChangedDate: $
           Revision: $Revision: $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2010
            License: GNU General Public License
-----------------------------------------------------
        This file is part of "DIMG".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

        File Notes:
        ------------------
*/

#ifndef FSFILE_H
#define FSFILE_H

#include <QSharedPointer>
#include <QScopedPointer>

#include "directoryentry.h"
#include "datachain.h"


namespace FilesystemFAT {

    class FsFile;
    class Directory;
    class FatFS;

    /** A shared pointer to a FsFile.
     */
    typedef QSharedPointer<FsFile> FsFilePtr;

    /** Represents a File in the FatFS Filesystem Implementation. It would have
        been called just "File" but that is a little too generic and likely to
        conflict with other things.

        It doesn't add much to DirectoryEntry besides proxying a few things
        through to an internal DataChain instance (seek(), pos(), length(),
        read(), readAll() and atEnd()) and a function to get the parent
        directory (which actually just uses the FatFS instance to do all the
        work).
     */
    class FsFile : public DirectoryEntry
    {
    Q_OBJECT
    public:
        /** Creates a new FsFile object.
          \param entryData The directory entry data. This is used to setup
                 attributes, filename, etc.
          \param filesystem The FatFS instance this FsFile is associated with.
          \param parentPathName The full name of the parent directory.
          \param parent The parent QObject. Should normally be the FatFS
                 instance.
         */
        explicit FsFile(QByteArray entryData,
                        FatFS *filesystem,
                        QString parentPathName,
                        QObject *parent = 0);

        /** Creates a new FsFile object. The longFilename is not stored in the
          directory entry data so it must be supplied separately.

          \param entryData The directory entry data. This is used to setup
                           attributes, filename, etc.
          \param filesystem The FatFS instance this FsFile is associated with.
          \param parentPathName The full name of the parent directory.
          \param longFilename The files Long Filename.
          \param parent The parent QObject. Should normally be the FatFS
                 instance.
         */
        explicit FsFile(QByteArray entryData,
                        FatFS *filesystem,
                        QString parentPathName,
                        QString longFilename,
                        QObject *parent);

        /** Creates a new instance of the parent directory.
          \returns A new instance of the parent directory (if it exists)
                   or NULL.
         */
        Directory *getParentDirectory();

        /* ***************************************************************** *
         * These functions just pass through to the internal DataChain       *
         * instance (dc) which does all the real work. They reset the        *
         * objects lastErr value to that supplied by the DataChain.          *
         * Hopefully this should make it appear as though FsFile directly    *
         * inherited from DataChain - something that is not at the moment    *
         * possible but may change sometime.                                 *
         * ***************************************************************** */

        /** \copydoc DataChain::seek()
         */
        bool seek(unsigned int pos);

        /** \copydoc DataChain::pos()
         */
        unsigned int pos();

        /** \copydoc DataChain::length()
         */
        unsigned int length();

        /** \copydoc DataChain::read()
         */
        QByteArray read(unsigned int length);

        /** \copydoc DataChain::readAll()
         */
        QByteArray readAll();

        /** \copydoc DataChain::atEnd()
         */
        bool atEnd();

    private:
        QScopedPointer<DataChain> dc;
        QPointer<FatFS> filesystem;
        QString parentPathName;
    };
}

#endif // FSFILE_H
