#include <QtCore/QCoreApplication>

#include <QTextStream>
#include <QFile>
#include <QList>
#include <QStringList>
#include <stdio.h>
#include <stdlib.h>
#include <QtDebug>
#include <QPointer>

#include "fatfs.h"
#include "biosparamblock.h"
#include "fsfile.h"
#include "directory.h"
#include "errors.h"
#include "util.h"


using namespace std;

/** Shows some brief help information.
 */
void showHelp();

/** Shows the contents of the BIOS Parameter Block.
  \param bpb The BIOS Parameter Block to show.
 */
void showBPB(FilesystemFAT::BiosParamBlock *bpb);

/** Prints out a directory listing for the supplied directory.
  \param dir The directory to print out to the console
 */
void printDirectory(FilesystemFAT::Directory *dir);

/** Prints a recursive directory listing starting in the supplied directory.
  \param startIn the directory to start in.
 */
void recursivePrintDirectory(FilesystemFAT::Directory *startIn);

/** Prints a recursive directory listing starting in the supplied directory.
  \param startIn the directory to start in.
 */
void recursivePrintDirectory(FilesystemFAT::DirectoryPtr startIn);

/** Builds a string representing a directory entry suitable for console output.
  \param dirent The directory entry to turn into a string.
  \returns A string representation of a directory entry.
 */
QString buildDirectoryEntryValue(FilesystemFAT::DirEntPtr dirent);

/** Extracts the specified file.
  \param filename The file to extract
  \param curdir The current directory
  \param fs The filesystem instance.
 */
void extractFile(QString filename,
                 FilesystemFAT::Directory* curdir,
                 FilesystemFAT::FatFS *fs);

/** Controls output of debug messages.
  \param type The message type.
  \param msg The message.
 */
void debugMessageOutput(QtMsgType type, const char *msg);

static bool dbgIncludeTypeInMessage;
static bool showDebugOutput;
static bool showWarnings;
static bool showErrors;

/** Test Program entry point. It does all the user input work and
  a great deal of the output work.
  \param argc Number of args.
  \param argv Args.
 */
int main(int argc, char *argv[])
{

    qInstallMsgHandler(debugMessageOutput);

    QCoreApplication a(argc, argv);

    // For console I/O
    QTextStream cin(stdin, QIODevice::ReadOnly);
    QTextStream cout(stdout, QIODevice::WriteOnly);
 //   QTextStream cerr(stderr, QIODevice::WriteOnly);


    dbgIncludeTypeInMessage = false;
    showDebugOutput = false;
    showWarnings = true;
    showErrors = true;
    bool longFilenamesEnabled = true;

    // The current disk image
    QFile imageFile;

    cout << "FatFS Test Program" << endl;
    cout << "    (C) Copyright David Goodwin, 2010\n" << endl;

    FilesystemFAT::FatFS *fs = NULL;
    FilesystemFAT::BiosParamBlock bpb;
     QPointer<FilesystemFAT::Directory> curdir = NULL;
    bool open = false;
    QString curpath = "/";

    QStringList autoexec;


    cout << "Type ? for help." << endl;
    while (true) {
        QString curlin;
        if (!autoexec.isEmpty()) {
            QString val = autoexec.takeFirst();
            cout << ">>> " + val << endl;
            cout.flush();
            curlin = val;
        } else {
            cout << ">>> ";
            cout.flush();
            curlin = cin.readLine();
        }

        QString compare = curlin.toUpper();
        if (compare == "EXIT" || compare == "QUIT") {
            return 0;
        } else if (compare.startsWith("HELP") ||
                   compare.startsWith("?")) {
            showHelp();
        } else if (compare.startsWith("OPEN ")) {

            if (open) {
                cout << "An image is already open. Close it first.";
            } else {

                QString filename = curlin.mid(5);
                cout << "Opening file '" << filename << "'" << endl;
                imageFile.setFileName(filename);
                if (!imageFile.open(QIODevice::ReadOnly)) {
                    cout << "Failed to open file." << endl;
                    continue;
                } else {
                    cout << "Image file open." << endl;
                }
                FilesystemFAT::Errors::ReturnCode ret = bpb.load(&imageFile);
                if (ret != FilesystemFAT::Errors::SUCCESS &&
                    ret != FilesystemFAT::Errors::BPB_WARNING) {
                    cout << "Error loading BPB: "
                         << FilesystemFAT::Errors::errorToString(ret) << endl;
                    open = false;
                    continue;
                }

                fs = new FilesystemFAT::FatFS(0);
                fs->setLFNSupportEnabled(longFilenamesEnabled);
                fs->setDevice(&imageFile);
                if (!fs->open()) {
                    ret = fs->lastError();
                    cout << "Failed to open Filesystem: "
                         << FilesystemFAT::Errors::errorToString(ret) << endl;
                } else {
                    curdir = fs->createDirectory(curpath);

                    ret = fs->lastError();
                    if (ret != FilesystemFAT::Errors::SUCCESS) {
                        cout << "Failed to get root directory." << endl;
                    }

                    cout << "Filesystem open." << endl;
                    cout << "Volume Label: " << fs->getVolumeLabel() << endl;
                }
                open = true;
            }
        } else if (compare.startsWith("CLOSE")) {
            delete fs;
            fs = NULL;
            bpb.close();
            imageFile.close();
            open = false;
            cout << "Filesystem closed." << endl;
        } else if (compare.startsWith("DIR")) {
            if (open) {
                if (compare == "DIR /R") {
                    if (curdir != NULL) {
                        recursivePrintDirectory(curdir);
                    } else {
                        cout << "No current directory.";
                    }
                } else {
                    if (curdir != NULL) {
                        printDirectory(curdir);
                    } else {
                        cout << "No current directory.";
                    }
                }
            } else {
                cout << "No filesystem open." << endl;
            }
        } else if (compare.startsWith("CD")) {
            QString tmpPath = curlin.mid(3);
            qDebug() << "CD into directory" << tmpPath << "requested.";
             QPointer<FilesystemFAT::Directory> tmpDir;
            FilesystemFAT::Errors::ReturnCode ret;

            if (curdir.isNull()) {
                qWarning() << "Current directory instance is NULL. This is "
                        "not good.";
                continue;
            }

            if (!tmpPath.startsWith("/")) {
                qDebug() << "Starting from current directory...";
                tmpDir = curdir->getDirectory(tmpPath);
                ret = curdir->lastError();
            }
            else {
                qDebug() << "Starting from FatFS instance...";
                tmpDir = fs->createDirectory(tmpPath);
                ret = fs->lastError();
            }

            if (ret != FilesystemFAT::Errors::SUCCESS) {
                cout << "CD Failed: "
                     << FilesystemFAT::Errors::errorToString(ret) << endl;
            } else if (tmpDir.isNull()) {
                qCritical() << "New directory is NULL.";
            } else {
                delete curdir;
                curpath = tmpPath;
                curdir = tmpDir;
                cout << "Now in: " + curdir->getFullPathName() << endl;
            }
        } else if (compare.startsWith("PWD")) {
            if (open) {
                if (curdir != NULL) {
                    curpath = curdir->getFullPathName();
                    cout << curpath << endl;
                } else {
                    cout << "No current directory.";
                }

            } else {
                cout << "No filesystem open." << endl;
            }
        } else if (compare.startsWith("EXTRACT ")) {
            if (open) {
                extractFile(compare.mid(8),curdir,fs);
            } else {
                cout << "No filesystem open." << endl;
            }
        } else if (compare.startsWith("SHOW")) {
            QString what = "";
            if (compare == "SHOW") {
                cout << "What: ";
                cout.flush();
                what = cin.readLine();
            } else {
                what = compare.mid(5);
            }
            what = what.toUpper();

            if (what == "BPB") {
                // Show the BIOS Parameter Block
                if (open)
                    showBPB(&bpb);
                else
                    cout << "Error: No open filesystem image." << endl;
            } else if (what == "SHOWDEBUG") {
                if (showDebugOutput)
                    cout << "ShowDebug is ON" << endl;
                else
                    cout << "ShowDebug is OFF" << endl;
            } else if (what == "SHOWWARNINGS") {
                if (showWarnings)
                    cout << "ShowWarnings is ON" << endl;
                else
                    cout << "ShowWarnings is OFF" << endl;

            } else if (what == "SHOWERRORS") {
                if (showErrors)
                    cout << "ShowErrors is ON" << endl;
                else
                    cout << "ShowErrors is OFF" << endl;

            } else if (what == "SHOWDBGLABELS") {
                if (dbgIncludeTypeInMessage)
                    cout << "ShowDbgLabels is ON" << endl;
                else
                    cout << "ShowDbgLabels is OFF" << endl;
            } else if (what == "LONGFILENAMES") {
                if (longFilenamesEnabled)
                    cout << "Long Filename support is ON" << endl;
                else
                    cout << "Long Filename support is OFF" << endl;
            } else {
                cout << "Error: unknown keyword '" << what << "'" << endl;
            }
        } else if (compare.startsWith("SET")) {
            QString what = "";
            if (compare == "SET") {
                cout << "What: ";
                cout.flush();
                what = cin.readLine();
            } else {
                what = compare.mid(4);
            }

            what = what.toUpper();

            if (what.startsWith("SHOWDEBUG ")) {
                if (what.startsWith("SHOWDEBUG ON")) {
                    showDebugOutput = true;
                } else if (what.startsWith("SHOWDEBUG OFF")) {
                    showDebugOutput = false;
                } else {
                    cout << "Unrecognised setting. Use ON or OFF.";
                }
            } else if (what.startsWith("SHOWWARNINGS ")) {
                if (what.startsWith("SHOWWARNINGS ON")) {
                    showWarnings = true;
                } else if (what.startsWith("SHOWWARNINGS OFF")) {
                    showWarnings = false;
                } else {
                    cout << "Unrecognised setting. Use ON or OFF.";
                }
            } else if (what.startsWith("SHOWERRORS ")) {
                if (what.startsWith("SHOWERRORS ON")) {
                    showErrors = true;
                } else if (what.startsWith("SHOWERRORS OFF")) {
                    showErrors = false;
                } else {
                    cout << "Unrecognised setting. Use ON or OFF.";
                }
            } else if (what.startsWith("SHOWDBGLABELS ")) {
                if (what.startsWith("SHOWDBGLABELS ON")) {
                    dbgIncludeTypeInMessage = true;
                } else if (what.startsWith("SHOWDBGLABELS OFF")) {
                    dbgIncludeTypeInMessage = false;
                } else {
                    cout << "Unrecognised setting. Use ON or OFF.";
                }
            } else if (what.startsWith("LONGFILENAMES ")) {
                if (what.startsWith("LONGFILENAMES ON")) {
                    longFilenamesEnabled = true;

                } else if (what.startsWith("LONGFILENAMES OFF")) {
                    longFilenamesEnabled = false;
                } else {
                    cout << "Unrecognised setting. Use ON or OFF.";
                }

                if (open)
                    fs->setLFNSupportEnabled(longFilenamesEnabled);
            }
        } else {
            cout << "Unknown Command '" << curlin << "'" << endl;
        }
    }

    return a.exec();
}

void showHelp() {
    QTextStream cout(stdout, QIODevice::WriteOnly);
    cout << "Help\n----" << endl;
    cout << "    EXIT\t\tExits the program" << endl;
    cout << "    OPEN [filename]\tOpens the specified file" << endl;
    cout << "    DIR\t\t\tLists the contents of the current directory" << endl;
    cout << "       /R\t\tRecursive directory listing" << endl;
    cout << "    PWD\t\t\tPrints the current working directory" << endl;
    cout << "    EXTRACT [filename]\tExtracts the specified file to the "
            "current directory" << endl;
    cout << "    SHOW [what]\t\tShows information about:" << endl;
    cout << "       BPB\t\tBIOS Parameter Block" << endl;
    cout << "       ShowDebug\tShows the current ShowDebug value" << endl;
    cout << "       ShowWarnings\tShows the current ShowWarnings value" << endl;
    cout << "       ShowErrors\tShows the current ShowErrows value" << endl;
    cout << "       ShowDbgLabels\tShows the current ShowDbgLabels "
            "value" << endl;
    cout << "       LongFilenames\tShows the current Long Filenames "
            "setting" << endl;
    cout << "    SET [value]\t\tSets a parameter:" << endl;
    cout << "       ShowDebug [ON|OFF]\tShows or hides debug messages" << endl;
    cout << "       ShowWarnings [ON|OFF]\tShows or hides warning "
            "messages" << endl;
    cout << "       ShowErrors [ON|OFF]\tShows or hides internal critical "
            "error messages" << endl;
    cout << "       ShowDbgLabels [ON|OFF]\tShows or hides message "
            "types" << endl;
    cout << "       LongFilenames [ON|OFF]\tEnables or disables long "
            "filename support" << endl;
}

void showBPB(FilesystemFAT::BiosParamBlock *bpb) {
    QTextStream cout(stdout, QIODevice::WriteOnly);

    // Process the Media Descriptor value a bit
    unsigned char md = bpb->mediaDescriptor();
    int md_i = (int)md;
    QString md_desc = "";
    switch (md) {
    case 0xF0:
        md_desc =                       "(3.5\" double sided, 80 tracks per "
                                        "side, 18 or 36\n"
                  "                      sectors per track (1.44MB or "
                  "2.88MB). 5.25\" double sided,\n"
                  "                      80 tracks per side, 15 sectors per "
                  "track (1.2MB). Also\n"
                  "                      used for other media types.)";
        break;
    case 0xF8:
        md_desc = "Fixed Disk";
        break;
    case 0xF9:
        md_desc =                       "(3.5\" Double sided, 80 tracks per "
                                        "side, 9 sectors\n"
                  "                      per track (720K). 5.25\" Double "
                  "Sided, 80 tracksper side,\n"
                  "                      15 sectors per track (1.2MB).)";
        break;
    case 0xFA:
        md_desc =                       "(5.25\" Single sided, 80 tracks per "
                                        "side, 8 sectors\n"
                  "                      per track (320K))";
        break;
    case 0xFB:
        md_desc =                       "(3.5\" Double sided, 80 tracks per "
                                        "side, 8 sectors\n"
                  "                      per track (640K))";
        break;
    case 0xFC:
        md_desc =                       "(5.25\" Single Sided, 40 tracks per "
                                        "side, 9 sectors\n"
                  "                      per track (180K))";
        break;
    case 0xFD:
        md_desc =                       "(5.25\" Double Sided, 40 tracks per "
                                        "side, 9 sectors\n"
                  "                      per track (360K). Also used for 8\")";
        break;
    case 0xFE:
        md_desc =                       "(5.25\" Single sided, 40 tracks per "
                                        "side, 8 sectors\n"
                  "                      per track (160K). Also used for 8\")";
        break;
    case 0xFF:
        md_desc =                       "(5.25\" Double sided, 40 tracks per "
                                        "side, 8 sectors\n"
                  "                     per track (320K).)";
        break;
    default:
        md_desc = "(Unrecognised value)";
        break;
    }

    int ext12pdn = (int)bpb->extPhysicalDriveNumber();
    int res = (int)bpb->ext12Reserved();
    int sig = (int)bpb->ext12ExtendedBootSignature();

    // Output everything
    cout << "BIOS Parameter Block\n--------------------" << endl;
    cout << "    Bytes Per Sector: " << bpb->bytesPerSector() << endl;
    cout << " Sectors Per Cluster: " << bpb->sectorsPerCluster() << endl;
    cout << "    Reserved Sectors: " << bpb->reservedSectors() << endl;
    cout << "      Number of FATs: " << bpb->numberOfFATs() << endl;
    cout << "Max Root Dir Entries: " << bpb->maxRootEntries() << endl;
    cout << "       Total Sectors: " << bpb->totalSectors() << endl;
    cout << "    Media Descriptor: 0x" << QString::number(md_i,16).toUpper()
         << " " << md_desc << endl;
    cout << "     Sectors Per FAT: " << bpb->sectorsPerFAT() << endl;
    cout << "   Sectors Per Track: " << bpb->sectorsPerTrack() << endl;
    cout << "     Number Of Heads: " << bpb->numberOfHeads() << endl;
    cout << "      Hidden Sectors: " << bpb->hiddenSectors() << endl;

    cout << "\nFAT12/FAT16 Extended BIOS Parameter Block" << endl;
    cout << "-----------------------------------------" << endl;
    cout << "  Physical Drive Number: " << ext12pdn << endl;
    cout << "  Reserved Sector Value: 0x" << QString::number(res,16) << endl;
    cout << "    Checkdisk Requested: "
         << (bpb->ext12ChkdskRequested() ? "True" : "False") << endl;
    cout << " Surface Scan Requested: "
         << (bpb->ext12SurfaceScanRequested() ? "True" : "False")
         << endl;
    cout << "Extended Boot Signature: 0x" << QString::number(sig,16) << endl;
    cout << "              Volume ID: " << bpb->ext12VolumeID() << endl;
    cout << "           Volume Label: '"
         << bpb->ext12VolumeLabel() << "'" << endl;
    cout << "    FAT Filesystem Type: '" << bpb->ext12FSType() << "'" << endl;

}

void printDirectory(FilesystemFAT::Directory *dir) {
    qDebug() << "Showing directory" << dir->fileName();
    QTextStream cout(stdout, QIODevice::WriteOnly);
    QList<FilesystemFAT::DirectoryPtr> dirs = dir->getSubdirs();
    QList<FilesystemFAT::FsFilePtr> files = dir->getFiles();

    for (int i = 0; i < dirs.count(); i++) {
        cout << buildDirectoryEntryValue(dirs.at(i));
    }

    for (int i = 0; i < files.count(); i++) {
        cout << buildDirectoryEntryValue(files.at(i));
    }
}

void recursivePrintDirectory(FilesystemFAT::Directory *startIn) {
    QTextStream cout(stdout, QIODevice::WriteOnly);
    cout << "--------------------------------------------------------" << endl;
    cout << "Directory: " << startIn->getFullPathName() << endl;
    cout << "--------------------------------------------------------" << endl;

    QList<FilesystemFAT::DirectoryPtr> dirs = startIn->getSubdirs();
    QList<FilesystemFAT::FsFilePtr> files = startIn->getFiles();

    for (int i = 0; i < dirs.count(); i++) {
        cout << buildDirectoryEntryValue(dirs.at(i));
    }

    for (int i = 0; i < files.count(); i++) {
        cout << buildDirectoryEntryValue(files.at(i));
    }

    cout << endl;
    for (int i = 0; i < dirs.count(); i++) {
        recursivePrintDirectory(dirs.at(i));
    }
}

void recursivePrintDirectory(FilesystemFAT::DirectoryPtr startIn) {
    QTextStream cout(stdout, QIODevice::WriteOnly);
    cout << "--------------------------------------------------------" << endl;
    cout << "Directory: " << startIn->getFullPathName() << endl;
    cout << "--------------------------------------------------------" << endl;

    QList<FilesystemFAT::DirectoryPtr> dirs = startIn->getSubdirs();
    QList<FilesystemFAT::FsFilePtr> files = startIn->getFiles();

    for (int i = 0; i < dirs.count(); i++) {
        cout << buildDirectoryEntryValue(dirs.at(i));
    }

    for (int i = 0; i < files.count(); i++) {
        cout << buildDirectoryEntryValue(files.at(i));
    }

    cout << endl;
    for (int i = 0; i < dirs.count(); i++) {
        recursivePrintDirectory(dirs.at(i));
    }
}

QString buildDirectoryEntryValue(FilesystemFAT::DirEntPtr dirent) {

    QString value;

    if (!dirent->isSubdirectory()) {
        value.append(dirent->fileName());
    } else {
        value.append("<");
        value.append(dirent->fileName());
        value.append(">");
    }

    if (value.length() < 8) value.append("\t");
    value.append("\t");

    // Do the attributes
    if (dirent->isReadOnly()) value.append("r");
    else value.append("-");
    if (dirent->isHidden()) value.append("h");
    else value.append("-");
    if (dirent->isSystem()) value.append("s");
    else value.append("-");
    if (dirent->isArchive()) value.append("a");
    else value.append("-");
    if (dirent->isVolumeLabel()) value.append("v");
    else value.append("-");
    if (dirent->isSubdirectory()) value.append("d");
    else value.append("-");

    // another layout tab
    value.append("\t");

    // and now for the file size.
    if (!dirent->isSubdirectory()) {
        QString sz = "B";
        int tfsz = dirent->size();
        if (tfsz > 1024) {
            tfsz = tfsz / 1024;
            sz = "KB";
        }
        if (tfsz > 1024) {
            tfsz = tfsz / 1024;
            sz = "MB";
        }
        if (tfsz > 1024) {
            tfsz = tfsz / 1024;
            sz = "GB";
        }
        value.append(QString::number(tfsz));
        value.append(sz);
    }

    value.append("\t");
    value.append(dirent->modified().toString());

    value.append("\t");
    value.append(QString::number(dirent->entryCluster()));

    return value + "\n";
}

void extractFile(QString filename,
                 FilesystemFAT::Directory* curdir,
                 FilesystemFAT::FatFS *fs) {
    QTextStream cout(stdout, QIODevice::WriteOnly);
    FilesystemFAT::FsFile *file;
    if (filename.startsWith("/")) {
        qDebug() << "Asking FS to create file";
        file = fs->createFile(filename);
        filename = file->fileName();
    } else {
        qDebug() << "Asking current directory to create file";
        file = curdir->getFile(filename);
    }
    if (file->lastError() != FilesystemFAT::Errors::SUCCESS) {
        cout <<  FilesystemFAT::Errors::errorToString(file->lastError())
             << endl;
        if (file != NULL) delete file;
        return;
    }

    QFile localFile(filename);
    if (!localFile.open(QIODevice::WriteOnly)) {
        cout << "Failed to open target file (" << filename
             << "). Extract aborted." << endl;
        delete file;
        return;
    }

    cout << "Extracting..." << endl;
    localFile.write(file->readAll());
    localFile.flush();
    localFile.close();

    cout << "Extract complete." << endl;

    delete file;
}

void debugMessageOutput(QtMsgType type, const char *msg)
 {
    if (dbgIncludeTypeInMessage) {
        switch (type) {
        case QtDebugMsg:
            if (showDebugOutput)
                fprintf(stderr, "Debug: %s\n", msg);
            break;
        case QtWarningMsg:
            if (showWarnings)
                fprintf(stderr, "Warning: %s\n", msg);
            break;
        case QtCriticalMsg:
            if (showErrors)
                fprintf(stderr, "Critical: %s\n", msg);
            break;
        case QtFatalMsg:
            fprintf(stderr, "Fatal: %s\n", msg);
            abort();
        }
    } else {
        switch (type) {
        case QtDebugMsg:
            if (showDebugOutput)
                fprintf(stderr, "%s\n", msg);
            break;
        case QtWarningMsg:
            if (showWarnings)
                fprintf(stderr, "%s\n", msg);
            break;
        case QtCriticalMsg:
            if (showErrors)
                fprintf(stderr, "%s\n", msg);
            break;
        case QtFatalMsg:
            fprintf(stderr, "Fatal: %s\n", msg);
            abort();
        }
    }
 }
