/*
-----------------------------------------------------
   Application Name: DIMG
          File Name: fatentry.h
       Last Changed: $LastChangedDate: 2010-04-04 00:18:51 +1300 (Sun, 04 Apr 2010) $
           Revision: $Revision: 199 $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2008,2009,2010
            License: GNU General Public License
   File Description: FATEntry Header
-----------------------------------------------------
        This file is part of DIMG.

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

File Notes:
------------------
*/

#ifndef FAT12ENTRY_H
#define FAT12ENTRY_H

// Qt headers
#include <QFileIconProvider>

// FSPlugin headers
#include <fsentryv2.h>

// Fat12fs headers
#include "fatfslib.h"
#include "FatFS/directoryentry.h"
#include "winutils.h"

using namespace FilesystemFAT;

class FatEntry : public FSEntry2
{
public:
    ~FatEntry() {}

    /** Constructs a new FatEntry from the supplied DirectoryEntry

      \param e The DirectoryEntry to construct this entry from
      \param directory The path for this entry. Eg, if the full filename is
               /tmp/file.txt, the directory would be /tmp
     */
    FatEntry(DirEntPtr e, QString directory);

    /** Returns this entries attributes.

      The attributes are defined in FSPlugin as a typedef for QFlags<Attribute>.

      \returns This entries attributes.
     */
    FSPlugin::Attributes getAttributes() const;

    /** Convenience function to see if this entry represents a directory. The
        same information can be obtained by looking at the value returned by
        getAttributes().

        \returns True if this entry represents a directory.
     */
    bool isDirectory() const;

    /** Returns the size of this entry. It has no meaning for directories so
      should return 0 if it is a directory.

      If there is any error, 0 should be returned.

      \returns The size of the file in Bytes.
     */
    quint64 getSize() const;

    /** Returns the base name of the file without the path or extension.

      The base name contains all characters in the filename up to (but not
      including) the first '.' character.

      \returns The base name of the entry.
     */
    QString baseName() const;

    /** Returns the filenames suffix. That is, every character after the first
      '.' character in the filename.

      \returns The suffix of the entry.
     */
    QString suffix() const;

    /** The full FileName of the entry.

      \returns The full filename of the entry.
     */
    QString fileName() const;

    /** Returns the file name, including the path.

      \returns The file name including the path.
     */
    QString filePath() const;

    /** Returns the file path.

      \Returns The file path.
     */
    QString path() const;

    /** Gets the specified Date value for the file.

      Only date values that the FSPlugins getFeatures() method has indicated
      as being supported should be requested here.

      The mapping is:
        - D_CreatedDate support is indicated by F_CreationDate
        - D_AccessedDate support is indicated by F_AccessDate
        - D_ModifiedDate support is indicated by F_ModDate


      \param dt The date type to return
      \returns The value for the specified DateType. If the specified DateType
            is for a time, QDate() will be returned.
     */
    QDate getDate(DateType dt) const;

    /** Gets the specified Time value for the file.

      Only date values that the FSPlugins getFeatures() method has indicated
      as being supported should be requested here.

      The mapping is:
        - D_CreatedTime support is indicated by F_CreationTime
        - D_AccessedTime support is indicated by F_AccessTime
        - D_ModifiedTime support is indicated by F_ModTime


      \param dt The time type to return
      \returns The value for the specified DateType. If the specified DateType
            is for a date, QTime() will be returned.
     */
    QTime getTime(DateType dt) const;

    /** Gets an icon that represents this files type for display purposes. How
        exactly this should be interpreted depends on the filesystems native
        platform. For example, FAT filesystems might be able to use
        QFileIconProvider to provide platform-native icons for a files
        extension.

      \returns An icon that represents this files type.
     */
    QIcon getIcon() const;

    // FSPlugin v2 functions to expose short filenames:

    /** Returns the alternate base name of the file without the path or
      extension.

      The alternate base name contains all characters in the filename up to
      (but not including) the first '.' character.

      The alternate names are used by filesystems supporting more than one
      namespace (in FatFS v2 the alternate names allow access to the MS-DOS
      filename).

      \returns The base name of the entries alternate filename.
     */
    virtual QString altBaseName() const;

    /** Returns the alternate filenames suffix. That is, every character after
      the first '.' character in the filename.

      The alternate names are used by filesystems supporting more than one
      namespace (in FatFS v2 the alternate names allow access to the MS-DOS
      filename).

      \returns The suffix of the entries alternate filename.
     */
    virtual QString altSuffix() const;

    /** The full alternate FileName of the entry.

      The alternate names are used by filesystems supporting more than one
      namespace (in FatFS v2 the alternate names allow access to the MS-DOS
      filename).

      \returns The full alternate filename of the entry.
     */
    virtual QString altFileName() const;

    /** Returns the alternate file ame, including the path.

      The alternate names are used by filesystems supporting more than one
      namespace (in FatFS v2 the alternate names allow access to the MS-DOS
      filename).

      \returns The alternate filename including the path.
     */
    virtual QString altFilePath() const;
private:
    DirEntPtr ent;
    QString entry_path;
};

#endif // FAT12ENTRY_H
