/*
-----------------------------------------------------
   Application Name: DIMG
          File Name: zipimagefile.cpp
         Start Date: 28/03/2010
       Last Changed: $LastChangedDate: 2011-06-12 16:41:13 +1200 (Sun, 12 Jun 2011) $
           Revision: $Revision: 310 $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2010
            License: GNU General Public License
   File Description: ZipImageFile implementation
-----------------------------------------------------
        This file is part of DIMG.

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

File Notes:
------------------
*/


#include "zipimagefile.h"

ZipImageFile::ZipImageFile(QObject *parent) :
    ImageFile(parent)
{
    buffer = 0;
    currentDevice = 0;
    errorCode = 0;
}

ZipImageFile::~ZipImageFile() {
    close();
}

void ZipImageFile::open(QString filename,
                        bool create) {
    qDebug() << "FPzip: Open requested for file: " << filename;

    if (create) {
        qDebug() << "FPzip: Requested open mode is CREATE. Will open the "
                "file and TRUNCATE if it exists.";
        close();

        qDebug() << "FPzip: Preparing new buffer";
        buffer = new QBuffer();
        buffer->open(QIODevice::ReadWrite);
        currentDevice = buffer;
        connect(currentDevice,
                SIGNAL(bytesWritten(qint64)),
                this,
                SLOT(bytesWritten(qint64)));
        this->filename = filename;

        imageName = QFileInfo(filename).baseName() + ".ima";

        cleanFile();

    } else {
        qDebug() << "FPzip: Requested open mode is EXISTING. Will open the "
                "file only if it exists. It will not be truncated.";
        openExisting(filename);
    }
}

void ZipImageFile::openExisting(QString filename) {
    qDebug() << "FPzip\\ZipImageFile: Attempting to open existing file: " << filename;
    close();

    // Figure out exactly what file we should be opening.
    QStringList fileList;
    getZipFileList(filename, &fileList);

    if (fileList.isEmpty()) return;

    qDebug() << "Searching for suitable image in file...";
    foreach (QString currentFileName, fileList) {
        if (QFileInfo(currentFileName).completeSuffix().toUpper() == "IMA") {
            imageName = currentFileName;
            break;
        }
    }
    qDebug() << "Chosen image is: " << imageName;

    close();

    qDebug() << "Preparing buffer...";
    buffer = new QBuffer();
    buffer->open(QIODevice::ReadWrite);
    currentDevice = buffer;

    QProgressDialog qpd;
    qpd.setWindowTitle("Loading...");

    currentDevice->seek(0);

    qDebug() << "Loading image from archive...";
    int ret = extractFileFromZipFile(filename,imageName,currentDevice,&qpd);
    if (ret != 0) {
        qWarning() << "Failed to open zip file. Aborting.";
        errorCode = ret;
        return;
    }

    comment = getZipFileComment(filename);

    connect(currentDevice,
            SIGNAL(bytesWritten(qint64)),
            this,
            SLOT(bytesWritten(qint64)));
    this->filename = filename;

    cleanFile();

    qDebug() << "done.";
}

QIODevice *ZipImageFile::device() {
    return currentDevice;

}

int ZipImageFile::save() {

    QProgressDialog qpd;
    qpd.setWindowTitle("Loading...");

    QScopedPointer<QSettings> settings;
    QString settingsFile;
    settingsFile = QCoreApplication::applicationName() + ".ini";
    qDebug() << "Checking for configuration file " << settingsFile;

    if (QFile::exists(settingsFile)) {
        // Load settings from there
        settings.reset(new QSettings(settingsFile, QSettings::IniFormat, 0));
    } else {
        settings.reset(new QSettings());
    }
    qDebug() << "Settings Loaded.";

    int compressionLevel = settings->value("FileHandling/CompressionLevel",6).toInt();
    bool compress = (compressionLevel == 0) ? false : true;

    currentDevice->seek(0);
    writeFileToZip(filename,                /* zipFileName */
                   currentDevice,           /* fileData */
                   imageName,               /* file filename */
                   QDateTime::currentDateTime(),    /* Date/time */
                   /* Setting a file-level comment seems to upset WinImage: */
                   "" ,                     /* file comment */
                   true,                    /* overwrite zip file if exists */
                   false,                   /* append zip file if exists */
                   compress,                /* compress file data */
                   compressionLevel,        /* compression level */
                   &qpd);                   /* progress dialog */

    setZipFileComment(filename, comment);

    cleanFile();
    return 0;
}

int ZipImageFile::close() {
    qDebug() << "FPzip/ZipImageFile: Attempting to close file";

    if (buffer) {
        delete buffer;
        buffer = 0;
    }
    currentDevice = 0;

    return 0;
}

ImageFile::Features ZipImageFile::getFeatures() const {
    return ImageFile::F_Cached
            | ImageFile::F_Floppy_Disk
            | ImageFile::F_Comment;
}

bool ZipImageFile::isOpen() const {
    return currentDevice->isOpen();
}

void ZipImageFile::bytesWritten(qint64 /*bytes*/) {
    dirtyFile();
}

void ZipImageFile::cleanFile() {
    dirty = false;
    qDebug() << "FPzip/ZipImageFile is now clean.";
}

void ZipImageFile::dirtyFile() {
    dirty = true;
    qDebug() << "FPzip/ZipImageFile is now dirty.";
}
