/*
-----------------------------------------------------
   Application Name: "DIMG"
Application Version: 1.1.0
          File Name: imagefile.h
         Start Date: 25/02/2010
       Last Changed: $LastChangedDate: 2011-04-09 00:17:11 +1200 (Sat, 09 Apr 2011) $
           Revision: $Revision: 295 $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2008, 2009, 2010
            License: GNU General Public License
   File Description: DIMG Class header
-----------------------------------------------------
        This file is part of "DIMG".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

        File Notes:
        ------------------
*/

#ifndef IMAGEFILE_H
#define IMAGEFILE_H

#include <QObject>
#include <QIODevice>

/** This class represents a Disk Image File (or a Tape Image I suppose - but
DIMG doesn't support those (yet)). Subclasses provide read/write access to
the disk image (what ever format it may be) through a QIODevice.
 */
class ImageFile : public QObject {
Q_OBJECT
public:

    enum Feature {
        F_NoFeatures    = 0x00000000,   /*!< No features are supported */
        F_Comment       = 0x00000001,   /*!< Can store a comment */
        F_RawFormat     = 0x00000002,   /*!< This ImageFile handles raw images */
        F_Cached        = 0x00000004,   /*!< This ImageFile is cached in memory. */
        F_Buffer        = 0x00000008,   /*!< This is just a buffer imagefile. */
        F_Floppy_Disk   = 0x00000010,   /*!< This is a floppy disk image. */
        F_ReadOnly      = 0x00000020    /*!< This image file is read-only. */
    };
    Q_DECLARE_FLAGS(Features,Feature)

    ImageFile(QObject *parent = 0): QObject(parent) {}

    /** Gets the filename this ImageFile represents
      \returns The filename this ImageFile represents.
     */
    virtual QString fileName() const = 0;

    /** Gets a QIODevice which can be used to read or write to the file this
      ImageFile represents
      \returns A QIODevice that can be used to access the file this ImageFile
      represents. This pointer belongs to the image file and should not be
      deleted.
     */
    virtual QIODevice *device() = 0;

    /** Saves the Image File if it is cached in memory. Otherwise does nothing.
      \returns An error code.
     */
    virtual int save() = 0;

    /** Closes the Image File, discarding any unsaved changes. The QIODevice will
      be closed and deleted at this time too.
      \returns An error code.
     */
    virtual int close() = 0;

    /** Gets features supported by this image file
      \returns Supported Features.
     */
    virtual Features getFeatures() const = 0;

    /** Gets the files comment (if supported - ask getFeatures() to see).
      \returns The files comment.
     */
    virtual QString getComment() const = 0;

    /** Sets the files comment (if supported - ask getFeatures() to see).
      \param comment The files new comment.
     */
    virtual void setComment(QString comment) = 0;

    /** Returns the maximum comment length.
      \returns The maximum comment length.
     */
    virtual long maxCommentLength() = 0;

    /** Returns true if the image is open and ready to go.
      \returns True if the image is open and ready to go.
     */
    virtual bool isOpen() const = 0;

    /** Returns true if the image has been changed since last save.
      \returns True if the image has been changes since last save.
      */
    virtual bool isDirty() const =0;
};

Q_DECLARE_OPERATORS_FOR_FLAGS(ImageFile::Features)

#endif // IMAGEFILE_H
