/*
-----------------------------------------------------
   Application Name: "DIMG"
Application Version: 1.1.0
          File Name: imagefileplugin.h
         Start Date: 25/02/2010
       Last Changed: $LastChangedDate: 2011-04-09 00:17:11 +1200 (Sat, 09 Apr 2011) $
           Revision: $Revision: 295 $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2008, 2009, 2010
            License: GNU General Public License
   File Description: ImageFilePlugin Plugin Interface
-----------------------------------------------------
        This file is part of "DIMG".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

        File Notes:
        ------------------
*/

#ifndef IMAGEFILEPLUGIN_H
#define IMAGEFILEPLUGIN_H

#include <QObject>
#include <QStringList>
#include <imagefile.h>

/** This is version 1 of the Image File Plugin interface for QtDIMG. It allows
  additional image file formats to be supported through the use of plugins.
 */
class ImageFilePlugin {

public:

    enum Feature {
        F_NoFeatures    = 0x00000000,   /*!< No features are supported         */
        F_RawFormat     = 0x00000001,   /*!< This ImageFile handles raw images.
                            fileIsSupported() will likely always return true.  */
        F_Read          = 0x00000002,   /*!< This image format can be read     */
        F_Write         = 0x00000004,   /*!< This image format can be written  */
        F_Comment       = 0x00000008,   /*!< Supports comments on some formats */
        F_FloppyDisk    = 0x00000010    /*!< This plugin is for floppy disks   */
    };
    Q_DECLARE_FLAGS(Features,Feature)

    enum LoadHint {
        LH_NoHints    = 0x00000000,   /*!< No load hints */
        LH_Create     = 0x00000001,   /*!< Create the image file if it does not exist */
        LH_NoCache    = 0x00000002,   /*!< Dont cache in memory */
    };
    Q_DECLARE_FLAGS(LoadHints,LoadHint)

    virtual ~ImageFilePlugin() {} /*!< Does Nothing. */

    /** Returns true if this ImageFilePlugin can load the specified
      file.
      \param filename The file to investigate.
      \returns True if this plugin is capable of loading the specified file.
     */
    virtual bool fileIsSupported(QString filename) = 0;

    /** Tries to open the specified file. Should it be unable to open
      the file, null is returned.

      This is a convenience function that does not specify any load hints.
      If the image file does not already exist it will not create it -
      instead null should be returned.

      \note This function can not be used to create new files.

      \param filename The file to try opening
      \returns An ImageFile instance representing that file (or null).
     */
    virtual ImageFile *createImageFile(QString filename) {
        return createImageFile(filename, LH_NoHints);
    }

    /** Tries to open the specified file using supplied load hints. Should it
      be unable to open the file null is returned.

      This is a convenience function that does not provide the ability to
      supply an expected size. When creating a new image file,
      createImageFile(filename,hints,expectedSize) should really be used.
      While this function can be used too, it omits the expectedSize paramter
      which will likely disable any in-memory caching of the image file (where
      supported) as the plugin can't possibly decide if it should cache it
      or not without knowing how big it will probably be.

      \param filename The file to try and open or create.
      \param hints Any load hints (such as create the file if it doesn't
             already exist on disk)
      \returns An ImageFile instance representing the on-disk file (or null
               in the case of an error).
     */
    virtual ImageFile *createImageFile(QString filename, LoadHints hints) {
        return createImageFile(filename,hints,0);
    }

    /** Tries to open the specified file using supplied load hints. Should it
      be unable to open the file null is returned.

      This function can take hints to assist in making decisions while loading
      the image file.

      \param filename The name of the file to try and open (or create)
      \param hints Any hints to assist in loading the image (such as
             instructing the plugin to create the file if it doesnt already
             exist)
      \param expectedSize The expected size of the image. This only needs to
             be supplied when
      \returns An ImageFile instance represting the on-disk file (or null
               in the case of an error).
     */
    virtual ImageFile *createImageFile(QString filename,
                                       LoadHints hints,
                                       long long expectedSize) = 0;

    /** Returns a list of filename filters that can be used in a file
      dialog.
      \returns A list of filename filters for use in a file dialog.
     */
    virtual QStringList getNameFilters() = 0;

    /** Gets the name of this ImageFilePlugin.
      \returns The plugins name.
     */
    virtual QString getName() const = 0;

    /** Gets a description of this ImageFilePlugin.
      \returns A description of this plugin.
     */
    virtual QString getDescription() const = 0;

    /** Gets a version string (like "1.0.3" or "0.2-1A") for this plugin.
      \returns This plugins version string.
     */
    virtual QString getVersion() const = 0;

    /** Gets this plugins copyright string.
      \returns The plugins copyright string.
    */
    virtual QString getCopyright() const { return "";}

    /** Returns the features supported by this plugin
      \returns The features supported by this plugin.
     */
    virtual Features getFeatures() const = 0;
};

Q_DECLARE_OPERATORS_FOR_FLAGS(ImageFilePlugin::Features)
Q_DECLARE_OPERATORS_FOR_FLAGS(ImageFilePlugin::LoadHints)

Q_DECLARE_INTERFACE(ImageFilePlugin,"nz.net.zx.DIMG.ImageFilePluginInterface/1.0")

#endif // IMAGEFILEPLUGIN_H
